import abc


from collections.abc import Callable


from watch import WatchMe
from watch.core import AttributeControllerMeta
from watch.builtins import InstanceOf, Container


class WatchABCMetaType(abc.ABCMeta, AttributeControllerMeta):
    pass


class WatchABCType(abc.ABC, WatchMe, metaclass=WatchABCMetaType):
    pass


class DAGNode(WatchABCType):
    data = InstanceOf(bytes)
    links = Container(InstanceOf(bytes), container=list)

    def __eq__(self, other):
        return (
            isinstance(self, type(other)) and
            self.data == other.data and
            self.links == other.links
        )

    def __init__(self, data, links):
        self.data = data
        self.links = links


class AbstractDriver(WatchABCType):
    """
    store  :: dag_node -> bytes
    lookup :: bytes -> dag_node
    """

    node_identifier = InstanceOf(Callable)

    def __init__(self, node_identifier):
        self.node_identifier = node_identifier

    @abc.abstractmethod
    def store(self, node: DAGNode) -> bytes:
        pass

    @abc.abstractmethod
    def lookup(self, node_id: bytes) -> DAGNode:
        pass




from sqlalchemy import Column, ForeignKey, Integer, LargeBinary


from sqlalchemy.ext.declarative import declarative_base


from sqlalchemy.orm import relationship


BaseMapping = declarative_base()


class Link(BaseMapping):
    __tablename__ = "links"
    id = Column(Integer, primary_key=True, autoincrement=True)
    parent = Column(LargeBinary(length=32), ForeignKey("nodes.id"), index=True)
    child = Column(LargeBinary(length=32), ForeignKey("nodes.id"), index=True)


class Node(BaseMapping):
    __tablename__ = "nodes"
    id = Column(LargeBinary(length=32), primary_key=True)
    data = Column(LargeBinary)
    links = relationship(Link, primaryjoin=Link.parent == id)




import click
import pathlib


from cadena.drivers.alchemy.helpers import (
    new_sqlite_driver_from_path, new_session_from_driver
)


from cadena.fslike_dag import Blob, Tree, LinkDescriptor, DATA, NAMESPACE


def store_dir(path, driver, session):
    tree = Tree.from_description(
        tree_type=NAMESPACE,
        link_descriptors=[
            LinkDescriptor(
                link_type=NAMESPACE if subpath.is_dir() else DATA,
                name=subpath.name,
                endpoint=store_path(subpath, driver, session)
            )
            for subpath in path.iterdir()
        ]
    )
    return driver.store(node=tree.dump(), session=session)


def store_file(path, driver, session, self_namespace=False):
    with open(str(path.absolute()), "rb") as f:
        blob_id = driver.store(
            node=Blob.from_data(data=f.read()).dump(),
            session=session
        )

    if not self_namespace:
        id_to_return = blob_id
    else:
        id_to_return = driver.store(
            node=Tree.from_description(
                tree_type=NAMESPACE,
                link_descriptors=[
                    LinkDescriptor(
                        endpoint=blob_id, name=path.name, link_type=DATA
                    )
                ]
            ).dump(),
            session=session
        )
    return id_to_return


def store_path(path, driver, session):
    if path.is_file():
        return store_file(path, driver, session)
    else:
        return store_dir(path, driver, session)


@click.command()
@click.argument("path_to_store", type=click.Path(exists=True))
@click.option("--parent", "-p", type=str, required=False)
def cli(path_to_store, parent):
    driver = new_sqlite_driver_from_path("snapshot.db")
    path = pathlib.Path(path_to_store)
    with new_session_from_driver(driver) as session:
        if path.is_file():
            root_id = store_file(path, driver, session, self_namespace=True)
        else:
            root_id = store_path(path=path, driver=driver, session=session)
        click.echo(root_id.hex())


if __name__ == "__main__":
    cli()




import watch


from cadena.abc import WatchABCType, DAGNode


from .proto import GenericNode, CommitData, TreeData, BlobData, LinkMeta
# pls dont remove next line
from .proto import NAMESPACE, DATA


def dump_to_dagnode(packed_node) -> DAGNode:
    return packed_node.dump()


def load_from_dagnode(dag_node: DAGNode):
    if dag_node is None:
        return None

    correspondance = {
        CommitData: Commit,
        TreeData: Tree,
        BlobData: Blob,
    }

    generic_node = GenericNode.FromString(dag_node.data)
    node_payload = getattr(generic_node, generic_node.WhichOneof("node_type"))

    return correspondance[type(node_payload)](
        packed_payload=node_payload, links=dag_node.links
    )


class _PBPackedNode(WatchABCType):

    links = watch.builtins.Container(
        items=watch.builtins.InstanceOf(bytes), container=list
    )

    def dump(self):
        return DAGNode(
            links=self.links,
            data=self.compose_generic_node().SerializeToString(),
        )

    def __init__(self, packed_payload, links):
        self.packed_payload = packed_payload
        self.links = links

    def __eq__(self, other):
        return (
            self.packed_payload == other.packed_payload and
            self.links == other.links
        )


class LinkDescriptor(watch.WatchMe):
    name = watch.builtins.InstanceOf(str) | watch.builtins.Just(None)
    endpoint = watch.builtins.InstanceOf(bytes)
    link_type = watch.builtins.Just(DATA, NAMESPACE)

    def __init__(self, endpoint, link_type, name=None):
        self.link_type = link_type
        self.name = name
        self.endpoint = endpoint


class Commit(_PBPackedNode):
    packed_payload = watch.builtins.InstanceOf(CommitData)

    def compose_generic_node(self):
        return GenericNode(commit=self.packed_payload)

    @property
    def tree(self) -> bytes:
        return self.links[0]

    @property
    def parents(self) -> list:
        return self.links[1:]

    @classmethod
    def from_tree_parents(cls, tree, parents):
        """
        tree::bytes: link to corresponding tree structure
        parents::list<bytes>: list of links to parent commits
        """
        return cls(
            packed_payload=CommitData(), links=[tree, *parents]
        )


class Tree(_PBPackedNode):
    packed_payload = watch.builtins.InstanceOf(TreeData)

    def compose_generic_node(self):
        return GenericNode(tree=self.packed_payload)

    @property
    def type(self):
        return self.packed_payload.type

    @classmethod
    def from_description(cls, tree_type, link_descriptors: list):
        # sanity check block
        if tree_type == DATA:
            all_fine = all(
                descriptor.name is None and descriptor.link_type == DATA
                for descriptor in link_descriptors
            )
        elif tree_type == NAMESPACE:
            all_fine = all(
                descriptor.name is not None for descriptor in link_descriptors
            )
        else:
            all_fine = False

        if not all_fine:
            raise ValueError("Misconfigured Tree node")

        return cls(
            packed_payload=TreeData(
                type=tree_type,
                mdata=[
                    LinkMeta(type=descriptor.link_type, name=descriptor.name)
                    for descriptor in link_descriptors
                ]
            ),
            links=[descriptor.endpoint for descriptor in link_descriptors]
        )


class Blob(_PBPackedNode):
    packed_payload = watch.builtins.InstanceOf(BlobData)

    @property
    def bytes(self):
        return self.packed_payload.data

    def compose_generic_node(self):
        return GenericNode(blob=self.packed_payload)

    @classmethod
    def from_data(cls, data: bytes):
        """
        data::bytes: binary string associated to the blob
        """
        return cls(
            packed_payload=BlobData(data=data),
            links=list(),
        )




# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dag.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='dag.proto',
  package='dag',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\tdag.proto\x12\x03\x64\x61g\"k\n\x07\x44\x41GNode\x12\x1d\n\x06\x63ommit\x18\x01 \x01(\x0b\x32\x0b.dag.CommitH\x00\x12\x19\n\x04tree\x18\x02 \x01(\x0b\x32\t.dag.TreeH\x00\x12\x19\n\x04\x62lob\x18\x03 \x01(\x0b\x32\t.dag.BlobH\x00\x42\x0b\n\tnode_type\"\x08\n\x06\x43ommit\"=\n\x10TreeLinkMetadata\x12\x1b\n\x04type\x18\x01 \x01(\x0e\x32\r.dag.NodeType\x12\x0c\n\x04name\x18\x02 \x01(\t\"I\n\x04Tree\x12\x1b\n\x04type\x18\x01 \x01(\x0e\x32\r.dag.NodeType\x12$\n\x05mdata\x18\x02 \x03(\x0b\x32\x15.dag.TreeLinkMetadata\"\x14\n\x04\x42lob\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c*#\n\x08NodeType\x12\r\n\tNAMESPACE\x10\x00\x12\x08\n\x04\x44\x41TA\x10\x01\x62\x06proto3')
)

_NODETYPE = _descriptor.EnumDescriptor(
  name='NodeType',
  full_name='dag.NodeType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NAMESPACE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATA', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=297,
  serialized_end=332,
)
_sym_db.RegisterEnumDescriptor(_NODETYPE)

NodeType = enum_type_wrapper.EnumTypeWrapper(_NODETYPE)
NAMESPACE = 0
DATA = 1



_DAGNODE = _descriptor.Descriptor(
  name='DAGNode',
  full_name='dag.DAGNode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='dag.DAGNode.commit', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tree', full_name='dag.DAGNode.tree', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blob', full_name='dag.DAGNode.blob', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='node_type', full_name='dag.DAGNode.node_type',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=18,
  serialized_end=125,
)


_COMMIT = _descriptor.Descriptor(
  name='Commit',
  full_name='dag.Commit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=127,
  serialized_end=135,
)


_TREELINKMETADATA = _descriptor.Descriptor(
  name='TreeLinkMetadata',
  full_name='dag.TreeLinkMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='dag.TreeLinkMetadata.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='dag.TreeLinkMetadata.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=137,
  serialized_end=198,
)


_TREE = _descriptor.Descriptor(
  name='Tree',
  full_name='dag.Tree',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='dag.Tree.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mdata', full_name='dag.Tree.mdata', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=200,
  serialized_end=273,
)


_BLOB = _descriptor.Descriptor(
  name='Blob',
  full_name='dag.Blob',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='dag.Blob.data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=275,
  serialized_end=295,
)

_DAGNODE.fields_by_name['commit'].message_type = _COMMIT
_DAGNODE.fields_by_name['tree'].message_type = _TREE
_DAGNODE.fields_by_name['blob'].message_type = _BLOB
_DAGNODE.oneofs_by_name['node_type'].fields.append(
  _DAGNODE.fields_by_name['commit'])
_DAGNODE.fields_by_name['commit'].containing_oneof = _DAGNODE.oneofs_by_name['node_type']
_DAGNODE.oneofs_by_name['node_type'].fields.append(
  _DAGNODE.fields_by_name['tree'])
_DAGNODE.fields_by_name['tree'].containing_oneof = _DAGNODE.oneofs_by_name['node_type']
_DAGNODE.oneofs_by_name['node_type'].fields.append(
  _DAGNODE.fields_by_name['blob'])
_DAGNODE.fields_by_name['blob'].containing_oneof = _DAGNODE.oneofs_by_name['node_type']
_TREELINKMETADATA.fields_by_name['type'].enum_type = _NODETYPE
_TREE.fields_by_name['type'].enum_type = _NODETYPE
_TREE.fields_by_name['mdata'].message_type = _TREELINKMETADATA
DESCRIPTOR.message_types_by_name['DAGNode'] = _DAGNODE
DESCRIPTOR.message_types_by_name['Commit'] = _COMMIT
DESCRIPTOR.message_types_by_name['TreeLinkMetadata'] = _TREELINKMETADATA
DESCRIPTOR.message_types_by_name['Tree'] = _TREE
DESCRIPTOR.message_types_by_name['Blob'] = _BLOB
DESCRIPTOR.enum_types_by_name['NodeType'] = _NODETYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DAGNode = _reflection.GeneratedProtocolMessageType('DAGNode', (_message.Message,), dict(
  DESCRIPTOR = _DAGNODE,
  __module__ = 'dag_pb2'
  # @@protoc_insertion_point(class_scope:dag.DAGNode)
  ))
_sym_db.RegisterMessage(DAGNode)

Commit = _reflection.GeneratedProtocolMessageType('Commit', (_message.Message,), dict(
  DESCRIPTOR = _COMMIT,
  __module__ = 'dag_pb2'
  # @@protoc_insertion_point(class_scope:dag.Commit)
  ))
_sym_db.RegisterMessage(Commit)

TreeLinkMetadata = _reflection.GeneratedProtocolMessageType('TreeLinkMetadata', (_message.Message,), dict(
  DESCRIPTOR = _TREELINKMETADATA,
  __module__ = 'dag_pb2'
  # @@protoc_insertion_point(class_scope:dag.TreeLinkMetadata)
  ))
_sym_db.RegisterMessage(TreeLinkMetadata)

Tree = _reflection.GeneratedProtocolMessageType('Tree', (_message.Message,), dict(
  DESCRIPTOR = _TREE,
  __module__ = 'dag_pb2'
  # @@protoc_insertion_point(class_scope:dag.Tree)
  ))
_sym_db.RegisterMessage(Tree)

Blob = _reflection.GeneratedProtocolMessageType('Blob', (_message.Message,), dict(
  DESCRIPTOR = _BLOB,
  __module__ = 'dag_pb2'
  # @@protoc_insertion_point(class_scope:dag.Blob)
  ))
_sym_db.RegisterMessage(Blob)


# @@protoc_insertion_point(module_scope)



from sqlalchemy import create_engine


from .driver import AlchemyDriver, new_session


def new_session_from_driver(driver):
    return new_session(driver.db_engine)


def new_sqlite_driver_from_path(path):
    return AlchemyDriver(db_engine=create_engine("sqlite:///%s" % path))


def new_sqlite_inmem_driver():
    return AlchemyDriver(db_engine=create_engine("sqlite:///:memory:"))




import json
import base64
import sys


import click


from cadena.drivers.alchemy.helpers import new_sqlite_driver_from_path
from cadena.fslike_dag import Commit, Blob, Tree, load_from_dagnode, NAMESPACE


def dump_node_stats(node):
    node_dict = {"type": type(node).__qualname__.upper()}

    if isinstance(node, Commit):
        node_dict["parents"] = [parent.hex() for parent in node.parents]
        node_dict["tree"] = node.tree.hex()

    elif isinstance(node, Tree):
        node_dict["type"] = (
            "NAMESPACE" if node.type == NAMESPACE else "DATA"
        )
        node_dict["subtrees"] = [
            {
                "id": link.hex(),
                "name": link_meta.name,
                "type": "NAMESPACE" if link_meta.type == NAMESPACE else "DATA"
            }
            for link_meta, link in zip(node.packed_payload.mdata, node.links)
        ]

    elif isinstance(node, Blob):
        node_dict["data"] = base64.b64encode(node.bytes).decode()

    return node_dict


@click.command()
@click.option("--db_path", type=click.Path(exists=True), default="snapshot.db")
@click.argument("tree_id", type=str)
def cli(db_path, tree_id):
    # try to parse tree_id
    try:
        tree_id_native = bytes.fromhex(tree_id)
    except ValueError:
        click.echo(
            "Fatal: %s is not a valid tree id." % tree_id,
            err=True, color="red"
        )
        sys.exit(1)

    driver = new_sqlite_driver_from_path(db_path)
    maybe_tree = load_from_dagnode(driver.lookup(node_id=tree_id_native))
    if maybe_tree is not None:
        click.echo(
            json.dumps(dump_node_stats(maybe_tree), indent=4, sort_keys=True)
        )


if __name__ == "__main__":
    cli()




from cadena.abc import AbstractDriver
from cadena.drivers.common import sha256_id


class MemdictDriver(AbstractDriver):

    def __init__(self):
        super().__init__(node_identifier=sha256_id)
        self.storage = dict()

    def store(self, node):
        node_id = self.node_identifier(node)
        self.storage[node_id] = node
        return node_id

    def lookup(self, node_id):
        return self.storage.get(node_id, None)




from contextlib import contextmanager


from sqlalchemy.orm import scoped_session, sessionmaker


from cadena.abc import AbstractDriver, DAGNode
from cadena.drivers.common import sha256_id


from .db_objects import BaseMapping, Node as DBNode, Link as DBLink


@contextmanager
def new_session(db_engine):
    session = scoped_session(sessionmaker(bind=db_engine))
    try:
        yield session
        session.commit()
    except Exception as e:
        session.rollback()
        raise e
    finally:
        session.close()


def autosession(method):
    """ Works with AlchemyDriver methods like "store" and "lookup".
    """

    def decorator(*args, **kwargs):
        self, *_ = args

        session = kwargs.get("session")
        if session is None:
            with new_session(self.db_engine) as s:
                kwargs["session"] = s
                return method(*args, **kwargs)
        return method(*args, **kwargs)

    return decorator


class AlchemyDriver(AbstractDriver):

    def __init__(self, db_engine):
        super().__init__(node_identifier=sha256_id)
        self.db_engine = db_engine
        BaseMapping.metadata.create_all(self.db_engine)

    @autosession
    def store(self, *, node, session=None):
        id_to_return = sha256_id(node)
        if self.lookup(node_id=id_to_return, session=session) is not None:
            return id_to_return

        session.add(
            DBNode(
                id=id_to_return,
                data=node.data,
                links=[
                    DBLink(child=link) for link in node.links
                ]
            )
        )

        return id_to_return

    @autosession
    def lookup(self, *, node_id, session=None):
        db_node = session.query(DBNode).filter_by(id=node_id).first()
        return (
            db_node and
            DAGNode(
                data=db_node.data,
                links=[
                    item.child for item in db_node.links
                ]
            )
        )




import time
import setuptools


BUILD_NUMBER = str(int(time.time()))


classifiers = [
    (
        "Programming Language :: Python :: %s" % x
    )
    for x in "3.5 3.6 3.7".split()
]


setuptools.setup(
    name="cadena",
    description="Wierd and inefficient Tar clone.",
    version="0.0.1.%s" % BUILD_NUMBER,
    license="MIT license",
    platforms=["unix", "linux", "osx", "win32"],
    author="magniff",
    url="https://github.com/magniff/cadena",
    classifiers=classifiers,
    install_requires=[
        "sqlalchemy", "watch", "protobuf", "click", "pymongo",
    ],
    entry_points={
        "console_scripts": [
            "store=bin.store:cli",
            "probe=bin.probe:cli",
        ]
    },
    packages=setuptools.find_packages(),
    zip_safe=False,
)




import pymongo


from cadena.abc import AbstractDriver, DAGNode
from cadena.drivers.common import sha256_id


class MongoDriver(AbstractDriver):

    def __init__(self, host="localhost", port=27017):
        super().__init__(node_identifier=sha256_id)
        self.storage = pymongo.MongoClient(host, port).cadena_db.cadena_data

    def store(self, node, *args, **kwargs):
        node_id = self.node_identifier(node)
        if self.storage.find_one({"id": node_id.hex()}):
            return node_id

        self.storage.insert_one(
            {
                "id": node_id.hex(),
                "data": node.data.hex(),
                "links": [link.hex() for link in node.links]
            }
        )

        return node_id

    def lookup(self, node_id, *args, **kwargs):
        result = self.storage.find_one({"id": node_id.hex()})
        return (
            result and
            DAGNode(
                data=bytes.fromhex(result["data"]),
                links=[
                    bytes.fromhex(link) for link in result["links"]
                ]
            )
        )

